__author__ = "Sivadon Chaisiri"
__copyright__ = "Copyright (c) 2020 Sivadon Chaisiri"
__license__ = "MIT License"


import sys
import json

import click

from . import pyhanga_constants as const
from botocore.exceptions import ClientError

def recaseTuple(torigin, ldesire):
    mUpperDesire = map(lambda x:x.upper(), ldesire)
    lUpperDesire = list(mUpperDesire)
    lorigin = list(torigin)
    for i in range(len(lorigin)):
        lorigin[i] = ldesire[lUpperDesire.index(lorigin[i].upper())]
    
    return tuple(lorigin)

def uppercaseTuple(torigin):
    mUpperOrigin = map(lambda x:x.upper(), torigin)
    return tuple(mUpperOrigin)

def reformS3Prefix(prefix):
    if prefix == '/':
        prefix = ''
    
    if prefix.startswith('/'):
        prefix = prefix[1:]
    
    if not prefix.endswith('/') and prefix != '':
        prefix = prefix + '/'
    
    return prefix


def getJsonDataFromFile(sJson, baseName):
    try:
        with open(sJson, 'r') as fJson:
            text = fJson.read()
            jsonData = json.loads(text)  
    except FileNotFoundError:
        handleFileNotFound(sJson)
    except json.decoder.JSONDecodeError: 
        click.secho(const.ERM_JSON_INVALID % baseName, bg=const.BG_ERROR, fg=const.FG_ERROR)
        sys.exit(const.ERC_FILE_NOTFOUND)    
    return jsonData                 

def handleFileNotFound(file):
        click.secho('%s: %s' % (const.ERM_FILE_NOTFOUND, file), bg=const.BG_ERROR, fg=const.FG_ERROR)
        sys.exit(const.ERC_FILE_NOTFOUND)    

def handleClientError(error):
    errorCode = str(error.response['Error']['Code'])
    errorMessage = str(error.response['Error']['Message'])
    showMessage = ''
    exitCode = 100
    
    if errorCode == 'ValidationError' and errorMessage.endswith('does not exist'):
        showMessage = 'The stack does not exist.'
    elif errorCode == 'ValidationError' and errorMessage.endswith('TerminationProtection is enabled'):
        showMessage = 'This stack cannot be deleted because it is protected with TerminationProtection.'
    elif errorCode == 'InsufficientCapabilitiesException' and errorMessage.endswith('[CAPABILITY_IAM]'):
        showMessage = 'This stack requires [CAPABILITY_IAM]. Specify --iam option to enable [CAPABILITY_IAM].'
    elif errorCode == 'InsufficientCapabilitiesException' and errorMessage.endswith('[CAPABILITY_NAMED_IAM]'):
        showMessage = 'This stack requires [CAPABILITY_NAMED_IAM]. Specify --iam option to enable [CAPABILITY_NAMED_IAM].'
    else:
        showMessage = error.response['Error']['Code'] + ' ' + error.response['Error']['Message'] 
                                
    click.secho('%s' % showMessage, bg=const.BG_ERROR, fg=const.FG_ERROR)   
    sys.exit(exitCode)


# Credit: http://code.activestate.com/recipes/577058/
def query_yes_no(question, default="no"):
    valid = {"yes": True, "y": True, "yes": True,
             "no": False, "n": False}
    if default == "yes":
        prompt = " [Y/n] "
    elif default == "no":
        prompt = " [y/N] "
    else:
        raise ValueError("invalid default answer: '%s'" % default)

    while True:
        sys.stdout.write(question + prompt)
        choice = input().lower()
        if choice in valid:
            return valid[choice]
        else:
            sys.stdout.write('Please respond with \'yes\' or \'no\'\n')






__author__ = "Sivadon Chaisiri"
__copyright__ = "Copyright (c) 2020 Sivadon Chaisiri"
__license__ = "MIT License"


import sys

import click
import boto3
import botocore

from . import _session
from . import pyhanga_constants as const
from . import pyhanga_util as util

@click.option('--name', '-n',
                required=True,
                help='Stack name')

@click.option('--field', '-f',
                help='Field to be printed out. You can print out one or more fields:\n'
                     'pyhanga dstack -f <field> -f <field> ...\n'
                     'By default, StackName, Description, StackStatus, CreateionTime, and EnableTerrminationProtection are printed out.',
                default = const.DEFAULT_DSTACK_FIELDS,
                multiple=True,
                type=click.Choice(const.STACK_DETAIL_FILEDS, case_sensitive=False))

@click.command(name='dstack')
def describe_stack(name, field):
    """
    Describe a stack
    """
    _describe_stack(name, field)

def _describe_stack(name, field = const.DEFAULT_DSTACK_FIELDS):
    try:
        response = _session.cf.describe_stacks(StackName=name)
    except botocore.exceptions.ClientError as error:
        util.handleClientError(error)
    # except:
    #     click.secho(const.ERM_OTHERS, bg=const.BG_ERROR, fg=const.FG_ERROR)
    #     sys.exit(const.ERC_OTHERS)  

    response = response[const.STACKS][0]

    field = util.recaseTuple(field, const.STACK_DETAIL_FILEDS)
    iField = iter(field)
    row = response[next(iField)]

    for i in iField:
        value = response.get(i)
        col = str(value) if value else const.NULL
        row = row + const.DELIM + col
    click.echo(row)
    return row 
    


__author__ = "Sivadon Chaisiri"
__copyright__ = "Copyright (c) 2020 Sivadon Chaisiri"
__license__ = "MIT License"


import sys
import os

import click
import boto3
import botocore

from . import _session
from . import pyhanga_constants as const
from . import pyhanga_util as util

@click.option('--bucket', '-b',
                help='S3 bucket',
                required=True,
                type=str)

@click.option('--prefix', '-p',
                help='S3 bucket prefix',
                default='/',
                required=True,
                type=str)

@click.option('--file', '-f',
                help='File to be uploaded',
                required=True,
                type=str)

@click.command(name='upload')
def upload_object(bucket, prefix, file):
    """
    Upload a file to a bucket
    """
    _upload_object(bucket, prefix, file)
 

def _upload_object(bucket, object_key, file):

    try:
        _session.s3.meta.client.upload_file(Filename=file, 
                                            Bucket=bucket, 
                                            Key=object_key)
    except FileNotFoundError:
        click.secho(const.ERM_FILE_NOTFOUND, bg=const.BG_ERROR, fg=const.FG_ERROR)
        sys.exit(const.ERC_FILE_NOTFOUND)   
    except botocore.exceptions.ClientError as error:
        util.handleClientError(error)
    # except:
    #     click.secho(const.ERM_OTHERS, bg=const.BG_ERROR, fg=const.FG_ERROR)
    #     sys.exit(const.ERC_OTHERS)

    return object_key 



__author__ = "Sivadon Chaisiri"
__copyright__ = "Copyright (c) 2020 Sivadon Chaisiri"
__license__ = "MIT License"


# AWS CLI 
DEFAULT_PROFILE = 'default' # Default profile
REGION_PROFILE = 'region' # Default region


# Stack fields and values - summary 
STACKS = 'Stacks'
STACK_SUMMARIES = 'StackSummaries'
STACK_NAME = 'StackName'
STACK_ID = 'StackId'
CHANGESET_ID = 'Id'
CREATION_TIME = 'CreationTime'
LAST_UPDATED_TIME = 'LastUpdatedTime'
DELETION_TIME = 'DeletionTime'
STACK_STATUS = 'StackStatus'
STACK_STATUS_REASON = 'StackStatusReason'
PARENT_ID = 'ParentId'
ROOT_ID = 'RootId'
DRIFT_INFORMATION ='DriftInformation'
ENABLE_TERMINATION_PROTECTION = 'EnableTerminationProtection'

# More stack fields
DESCRIPTION = 'Description'
TIMEOUT_IN_MINUTES = 'TimeoutInMinutes'
ROLE_ARN = 'RoleARN'
PARAMETERS = 'Parameters'
TAGS = 'Tags'
ENABLE_TERMINATION_PROTECTION= 'EnableTerminationProtection'
OUTPUTS = 'Outputs'
CAPABILITIES = 'Capabilities'
NOTIFICATION_ARNS = 'NotificationARNs'
DISABLE_ROLLBACK = 'DisableRollback'
ROLLBACK_CONFIGURATION = 'RollbackConfiguration'


STACK_SUMMARY_FILEDS = [STACK_NAME,
                    STACK_ID, 
                    CREATION_TIME,
                    LAST_UPDATED_TIME,
                    DELETION_TIME,
                    STACK_STATUS,
                    STACK_STATUS_REASON,
                    PARENT_ID,
                    ROOT_ID,
                    DRIFT_INFORMATION]

STACK_DETAIL_FILEDS = STACK_SUMMARY_FILEDS + [DESCRIPTION,
                                        TIMEOUT_IN_MINUTES,
                                        ROLE_ARN,
                                        PARAMETERS,
                                        TAGS,
                                        ENABLE_TERMINATION_PROTECTION,
                                        OUTPUTS,
                                        CAPABILITIES,
                                        NOTIFICATION_ARNS,
                                        DISABLE_ROLLBACK,
                                        ROLLBACK_CONFIGURATION]

# Stack status filters
CREATE_IN_PROGRESS = 'CREATE_IN_PROGRESS'
CREATE_FAILED = 'CREATE_FAILED'
CREATE_COMPLETE = 'CREATE_COMPLETE'
ROLLBACK_IN_PROGRESS = 'ROLLBACK_IN_PROGRESS'
ROLLBACK_FAILED = 'ROLLBACK_FAILED'
ROLLBACK_COMPLETE = 'ROLLBACK_COMPLETE'
DELETE_IN_PROGRESS = 'DELETE_IN_PROGRESS'
DELETE_FAILED = 'DELETE_FAILED'
DELETE_COMPLETE = 'DELETE_COMPLETE'
UPDATE_IN_PROGRESS = 'UPDATE_IN_PROGRESS'
UPDATE_COMPLETE_CLEANUP_IN_PROGRESS = 'UPDATE_COMPLETE_CLEANUP_IN_PROGRESS'
UPDATE_COMPLETE = 'UPDATE_COMPLETE'
UPDATE_ROLLBACK_IN_PROGRESS = 'UPDATE_ROLLBACK_IN_PROGRESS'
UPDATE_ROLLBACK_FAILED = 'UPDATE_ROLLBACK_FAILED'
UPDATE_ROLLBACK_COMPLETE_CLEANUP_IN_PROGRESS = 'UPDATE_ROLLBACK_COMPLETE_CLEANUP_IN_PROGRESS'
UPDATE_ROLLBACK_COMPLETE = 'UPDATE_ROLLBACK_COMPLETE'
REVIEW_IN_PROGRESS = 'REVIEW_IN_PROGRESS'
IMPORT_IN_PROGRESS = 'IMPORT_IN_PROGRESS'
IMPORT_COMPLETE = 'IMPORT_COMPLETE'
IMPORT_ROLLBACK_IN_PROGRESS = 'IMPORT_ROLLBACK_IN_PROGRESS'
IMPORT_ROLLBACK_FAILED = 'IMPORT_ROLLBACK_FAILED'
IMPORT_ROLLBACK_COMPLETE = 'IMPORT_ROLLBACK_COMPLETE'



# Filter types
EXACTLY = 'exactly'
CONTAINS = "contains"
STARTS_WITH = "startswith"
ENDS_WITH = "endswith"

STRING_MATCH_CONDITIONS = [
                        EXACTLY,
                        CONTAINS,
                        STARTS_WITH,
                        ENDS_WITH]

# change set fields and status
CS_STATUS = 'Status'
CS_EXECUTION_STATUS = 'ExecutionStatus'
CS_CREATE_COMPLETE = 'CREATE_COMPLETE'
CS_FAILED = 'FAILED'
CS_UNAVAILABLE = 'UNAVAILABLE'
CS_AVAILABLE = 'AVAILABLE'
CS_EXECUTE_IN_PROGRESS = 'EXECUTE_IN_PROGRESS'
CS_EXECUTE_COMPLETE = 'EXECUTE_COMPLETE'
CS_EXECUTE_FAILED = 'EXECUTE_FAILED'
CS_OBSOLETE = 'OBSOLETE',
CS_RESOURCE_CHANGE = 'ResourceChange'
CS_CHANGES = 'Changes'
CS_ACTION = 'Action'
CS_ADD = 'Add'
CS_MODIFY = 'Modify'
CS_REMOVE = 'Remove'
CS_IMPORT = 'Import'
CS_REPLACEMENT = 'Replacement'

# Stack events
STACK_EVENTS = 'StackEvents'
EVENT_ID = 'EventId'
LOGICAL_RESOURCE_ID = 'LogicalResourceId'
PHYSICAL_RESOURCE_ID = 'PhysicalResourceId'
RESOURCE_TYPE = 'ResourceType'
TIMESTAMP = 'Timestamp'
RESOURCE_STATUS = 'ResourceStatus'
RESOURCE_STATUS_REASON = 'ResourceStatusReason'

# Stack resources
STACK_RESOURCE_SUMMARIES = 'StackResourceSummaries'
STACK_RESOURCE_DETAIL = 'StackResourceDetail'
LAST_UPDATED_TIMESTAMP = 'LastUpdatedTimestamp'
META_DATA = 'Metadata'
DRIFT_INFORMATION = 'DriftInformation'

# Default collections

STACK_STATUS_FILTERS = [
    CREATE_IN_PROGRESS,
    CREATE_FAILED,
    CREATE_COMPLETE,
    ROLLBACK_IN_PROGRESS,
    ROLLBACK_FAILED,
    ROLLBACK_COMPLETE,
    DELETE_IN_PROGRESS,
    DELETE_FAILED,
    DELETE_COMPLETE,
    UPDATE_IN_PROGRESS,
    UPDATE_COMPLETE_CLEANUP_IN_PROGRESS,
    UPDATE_COMPLETE,
    UPDATE_ROLLBACK_IN_PROGRESS,
    UPDATE_ROLLBACK_FAILED,
    UPDATE_ROLLBACK_COMPLETE_CLEANUP_IN_PROGRESS,
    UPDATE_ROLLBACK_COMPLETE,
    REVIEW_IN_PROGRESS,
    IMPORT_IN_PROGRESS,
    IMPORT_COMPLETE,
    IMPORT_ROLLBACK_IN_PROGRESS,
    IMPORT_ROLLBACK_FAILED,
    IMPORT_ROLLBACK_COMPLETE
]

STACK_STATUS_FILTERS_NO_DELETE_COMPLETE = STACK_STATUS_FILTERS[:]
STACK_STATUS_FILTERS_NO_DELETE_COMPLETE.remove(DELETE_COMPLETE)

STACK_EVENT_FIELDS = [EVENT_ID,
                RESOURCE_TYPE,
                LOGICAL_RESOURCE_ID,
                PHYSICAL_RESOURCE_ID,
                TIMESTAMP,
                RESOURCE_STATUS,
                RESOURCE_STATUS_REASON]

STACK_RESOURCE_FILEDS = [RESOURCE_TYPE,
                            LOGICAL_RESOURCE_ID,
                            PHYSICAL_RESOURCE_ID,
                            LAST_UPDATED_TIMESTAMP,
                            RESOURCE_STATUS,
                            RESOURCE_STATUS_REASON,
                            DESCRIPTION,
                            META_DATA,
                            DRIFT_INFORMATION]

STACK_RESOURCE_SUMMARY_FILEDS = STACK_RESOURCE_FILEDS[:] 
STACK_RESOURCE_SUMMARY_FILEDS.remove(DESCRIPTION)
STACK_RESOURCE_SUMMARY_FILEDS.remove(META_DATA)

DEFAULT_LSTACKSTATUS_FIELDS = tuple(STACK_STATUS_FILTERS_NO_DELETE_COMPLETE)

DEFAULT_DSTACK_FIELDS = tuple([STACK_NAME, 
                            DESCRIPTION,
                            STACK_STATUS, 
                            CREATION_TIME,
                            ENABLE_TERMINATION_PROTECTION])

DEFAULT_DEVENTS_FIELDS = tuple([RESOURCE_TYPE,
                                LOGICAL_RESOURCE_ID,
                                RESOURCE_STATUS,
                                TIMESTAMP])

DEFAULT_RESOURCE_FIELDS = tuple([RESOURCE_TYPE,
                                LOGICAL_RESOURCE_ID,
                                RESOURCE_STATUS,
                                LAST_UPDATED_TIMESTAMP])


# Capabilities
CAPABILITY_IAM = 'CAPABILITY_IAM'
CAPABILITY_NAMED_IAM = 'CAPABILITY_NAMED_IAM'

# Errror codes
ERC_PROFILE_NOTFOUND = 1
ERC_FILE_NOTFOUND = 2
ERC_JSON_INVALID = 3
ERC_S3_INVALID = 5
ERC_OTHERS = 100

# Error message
ERM_PROFILE_NOTFOUND = "Profile not found or invalid endpoint/region."
ERM_FILE_NOTFOUND = "Local file not found"
ERM_JSON_INVALID = "Invalid JSON: %s"
ERM_S3_INVALID = "S3 not found or not accessible"
ERM_OTHERS = "Error catched!"

# Informatio messages

INF_NOTHING_TO_CHANGE = 'Nothing to change.'

# Colors
FG_WARN = 'red'
FG_ERROR = 'red'
BG_ERROR = 'black'
FG_INF = 'green'

# File extensions
YAML_EXT = '.yaml'
JSON_EXT = '.json'

# Others
ALL = '*'
DELIM = ','
NULL = 'NULL'
ANIM_STRING = '|/-\\'
ANIM_LEN = len(ANIM_STRING)

DELAY_TIME_FOR_DESCRIBE_CHANGE_SET = 5
DELAY_TIME_FOR_ANIMATION = 0.25

NEXT_TOKEN = 'NextToken'

URL = 'Url'

MAX_RESOURCES_STACK = 200


__author__ = "Sivadon Chaisiri"
__copyright__ = "Copyright (c) 2020 Sivadon Chaisiri"
__license__ = "MIT License"

import sys

import click
import boto3
import botocore

from . import _session
from . import pyhanga_constants as const
from . import pyhanga_util as util

@click.option('--name', '-n',
                required=True,
                help='Stack name')

@click.option('--field', '-f',
                help='Field to be printed out. You can print out one or more fields:\n'
                     'pyhanga devents -f <field> -f <field> ...\n'
                     'By default, ResourceType, LogicalResourceId, ResourceStatus, and Timestamp are printed out.',
                default = const.DEFAULT_DEVENTS_FIELDS,
                multiple=True,
                type=click.Choice(const.STACK_EVENT_FIELDS, case_sensitive=False))

@click.option('--max-items', '-m',
                help='Maximum number of events to be returned (default: 1)\n',
                default = 1,
                type=int)

@click.command(name='devents')
def describe_events(name, field, max_items):
    """
    Describe events of a stack
    """
    _describe_events(name, field, max_items)

def _describe_events(name, field, max_items):
     nextToken = 'null'
     count_results = 0
     
     while True and count_results < max_items:
          try:
               response = _session.cf.describe_stack_events(StackName=name, NextToken=nextToken)
          except botocore.exceptions.ClientError as error:
               util.handleClientError(error)
          # except:
          #     click.secho(const.ERM_OTHERS, bg=const.BG_ERROR, fg=const.FG_ERROR)
          #     sys.exit(const.ERC_OTHERS)  
          
          field = util.recaseTuple(field, const.STACK_EVENT_FIELDS)
          
          json = response[const.STACK_EVENTS]

          for event in json:
               if count_results >= max_items:
                    break

               iField = iter(field)
               row = event[next(iField)]

               for i in iField:
                    value = event.get(i)
                    col = str(value) if value else const.NULL
                    row = row + const.DELIM + col 
               count_results += 1
               click.echo(row)

          nextToken = response.get(const.NEXT_TOKEN)
          if nextToken is None:
               break
     
     click.secho('\nTotal number of events returned:', fg=const.FG_INF)
     click.echo(count_results)


__author__ = "Sivadon Chaisiri"
__copyright__ = "Copyright (c) 2020 Sivadon Chaisiri"
__license__ = "MIT License"

import sys

import click
import boto3
import botocore

from . import _session
from . import pyhanga_constants as const
from . import pyhanga_util as util

@click.option('--name', '-n',
                required=True,
                help='Stack name')

@click.option('--lid', '-l',
                required=True,
                help='Logical resource ID')

@click.option('--field', '-f',
                help='Field to be printed out. You can print out one or more fields:\n'
                     'pyhanga dresource -f <field> -f <field> ...\n'
                     'By default, ResourceType, LogicalResourceId, ResourceStatus, and Timestamp are printed out.',
                default = const.DEFAULT_RESOURCE_FIELDS,
                multiple=True,
                type=click.Choice(const.STACK_RESOURCE_FILEDS, case_sensitive=False))


@click.command(name='dresource')
def describe_resource(name, lid, field):
    """
    Describe a resource of a stack
    """
    _describe_resource(name, lid, field)

def _describe_resource(name, lid, field):
     try:
          response = _session.cf.describe_stack_resource(StackName=name, 
                                                            LogicalResourceId=lid)
     except botocore.exceptions.ClientError as error:
          util.handleClientError(error)
     # except:
     #     click.secho(const.ERM_OTHERS, bg=const.BG_ERROR, fg=const.FG_ERROR)
     #     sys.exit(const.ERC_OTHERS)  
          
     field = util.recaseTuple(field, const.STACK_RESOURCE_FILEDS)
     
     resource = response[const.STACK_RESOURCE_DETAIL]
     iField = iter(field)
     row = resource[next(iField)]

     for i in iField:
          value = resource.get(i)
          col = str(value) if value else const.NULL
          row = row + const.DELIM + col 
     click.echo(row)



__author__ = "Sivadon Chaisiri"
__copyright__ = "Copyright (c) 2020 Sivadon Chaisiri"
__license__ = "MIT License"


# internal packages
import sys

# 3rd-party packages
import click
import botocore

# custom packages
from . import _session
from . import list_stacks
from . import list_resources
from . import describe_stack
from . import describe_events
from . import describe_resource
from . import create_stack
from . import update_stack
from . import delete_stack
from . import protect_stack
from . import cost_template
from . import upload_object
from . import pyhanga_util as util
from . import pyhanga_constants as const


def _print_version(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo('hanga 0.01')
    ctx.exit()

def _init_profile(ctx, param, value):
    click.echo(value)
    ctx.exit()


@click.group()
@click.option(
    '--version',
    is_flag=True,
    callback=_print_version,
    expose_value=False,
    is_eager=True,
    help='Show pyhanga version'
)

@click.option(
    '--profile', '-p',
    required=True,
    default=const.DEFAULT_PROFILE,
    help='AWS CLI profile'
)

@click.option(
    '--region', '-r',
    help='Working region'
)

def cli(profile, region):
    try:
        _session._init_session(profile, region)  
    except:
        click.secho(const.ERM_PROFILE_NOTFOUND, bg=const.BG_ERROR, fg=const.FG_ERROR)
        sys.exit(const.ERC_PROFILE_NOTFOUND)


cli.add_command(describe_stack.describe_stack)
cli.add_command(describe_events.describe_events)
cli.add_command(describe_resource.describe_resource)
cli.add_command(list_stacks.list_stacks)
cli.add_command(list_resources.list_resources)
cli.add_command(create_stack.create_stack)
cli.add_command(update_stack.update_stack)
cli.add_command(delete_stack.delete_stack)
cli.add_command(protect_stack.protect_stack)
cli.add_command(cost_template.cost_template)
cli.add_command(upload_object.upload_object)









__author__ = "Sivadon Chaisiri"
__copyright__ = "Copyright (c) 2020 Sivadon Chaisiri"
__license__ = "MIT License"

import boto3
import click

def _init_session(profile, region):
    session = boto3.Session(profile_name=profile,
                            region_name=region)

    global cf
    cf = session.client('cloudformation')

    global s3
    s3 = session.resource('s3')




__author__ = "Sivadon Chaisiri"
__copyright__ = "Copyright (c) 2020 Sivadon Chaisiri"
__license__ = "MIT License"
import click

from . import deploy_stack

@click.option('--name', '-n',
                required=True,
                type=click.STRING,
                help='Stack name')
                
@click.option('--template', '-t',
                help='Stack template',
                type=click.STRING)

@click.option('--bucket', '-b',
                help='S3 bucket',
                required=True)

@click.option('--object-prefix', '-o',
                help='S3 object prefix (i.e., directory)\n'
                        'By default, the prefix is empty (i.e., root the bucket).',
                default='')

@click.option('--params', '--parameters', 
                help='Parameter file',
                default=None)

@click.option('--tags',
                help='Tag file',
                default=None)               

@click.option('--upload', '-u',
                help='Upload the template file to the bucket prefix',
                default=False,
                is_flag=True)  

@click.option('--yes', '-y',
                help='(NOT RECOMMEND!) Proceed the stack creation without prompting',
                is_flag=True)  

@click.option('--iam',
                help='Enable CAPABILITY_IAM',
                is_flag=True)  

@click.option('--named-iam',
                help='Enable CAPABILITY_NAMED_IAM',               
                default=False,
                is_flag=True)  

# @click.option('--auto-expand',
#                 help='Enable CAPABILITY_AUTO_EXPAND',
#                 default=False,
#                 is_flag=True)  

@click.option('--default', '-d',
                help='Default files to be deployed, that is,\n'
                     '  {stack name}.yaml - Template file\n'
                     '  {stack name}-params.json - parameter file\n'
                     '  {stack name}-tags.json - tag file.\n',
                default=False,
                is_flag=True)                               

@click.command(name='create')
def create_stack(name, template, bucket, object_prefix, params, tags, upload, yes, iam, named_iam, default):
    """
    Create a new stack with a change set
    """
    deploy_stack.deploy_stack(name, template, bucket, False, object_prefix, params, tags, upload, yes, iam, named_iam, default, False)    



__author__ = "Sivadon Chaisiri"
__copyright__ = "Copyright (c) 2020 Sivadon Chaisiri"
__license__ = "MIT License"


import sys

import click
import boto3
import botocore

from . import _session
from . import pyhanga_constants as const
from . import pyhanga_util as util
from .describe_stack import _describe_stack

@click.option('--name', '-n',
                required=True,
                help='Stack name\n'
                        'If only this option is used, the command shows the current termination protection status of the stack.')

@click.option('--enable/--disable', '-e/-d',
                help='Enable/disable the termination protection',
                default=None,
                is_flag=True)             

@click.command(name='protect')
def protect_stack(name, enable):
    """
    Enable a stack to be protected from termination 
    """    

    click.secho('Current termination protection status: ', fg='green')  
    status = str(_describe_stack(name, 
            tuple([const.ENABLE_TERMINATION_PROTECTION])))

    if enable is None:
        click.secho(const.INF_NOTHING_TO_CHANGE, fg=const.FG_INF)
    else:  
        senable = str(enable)
        
        if (status in senable):
            click.secho(const.INF_NOTHING_TO_CHANGE, fg=const.FG_INF)
        else:
            try:
                _session.cf.update_termination_protection(StackName=name,
                                                        EnableTerminationProtection=enable)
                click.secho('The new termination protection status:', fg='green')
                click.echo(enable)
            except botocore.exceptions.ClientError as error:
                util.handleClientError(error)
            # except:
            #     click.secho(const.ERM_OTHERS, bg=const.BG_ERROR, fg=const.FG_ERROR)
            #     sys.exit(const.ERC_OTHERS)      


__author__ = "Sivadon Chaisiri"
__copyright__ = "Copyright (c) 2020 Sivadon Chaisiri"
__license__ = "MIT License"

import sys
import os
import random
import time

import click
import boto3
import botocore

from . import _session
from . import pyhanga_constants as const
from . import pyhanga_util as util
from .list_resources import _list_resources

@click.option('--name', '-n',
                required=True,
                type=click.STRING,
                help='Stack name')

@click.option('--yes', '-y',
                help='(NOT RECOMMEND!) Proceed the stack update without prompting',
                is_flag=True)  

@click.command(name='delete')
def delete_stack(name, yes):
    """
    Delete a stack
    """    
    try:
        deleteAction = False

        _list_resources(name, 
                        [const.RESOURCE_TYPE, const.LOGICAL_RESOURCE_ID],
                        const.MAX_RESOURCES_STACK)

        if not yes:
            click.secho('WARNING: The above resources of the stack will be deleted and unrecoverable if they are unprotected with RETAIN.', fg=const.FG_WARN)
        
        # If --yes flag is not specified, prompt the confirmation question.
        if not yes: 
            yes = util.query_yes_no('\nDo you want to execute this change set?', 'no')

        if not yes:
            click.secho('You have aborted this change set.', fg=const.FG_INF) 
        else:
            _session.cf.delete_stack(StackName=name)
            deleteAction = True
            eResponse = _wait_for_done_deletion(name)
            if eResponse != const.DELETE_COMPLETE:
                click.secho('The stack was unsuccessfully deleted with this status: %s!' % eResponse, fg=const.FG_ERROR)
            else:
                click.secho('The stack has been deleted!', fg=const.FG_INF)                
    except botocore.exceptions.ClientError as error:
        if deleteAction and str(error.response['Error']['Code']) == 'ValidationError' and str(error.response['Error']['Message']).endswith('does not exist'):
            click.secho('The stack has been deleted!', fg=const.FG_INF)  
        else:
            util.handleClientError(error)

def _wait_for_done_deletion(name):  
    click.secho('Please wait while the stack is being deleted...', fg=const.FG_INF)

    eResponse = const.DELETE_IN_PROGRESS
    response = None   
    running_time = 0
    anim_index = 0
    
    while eResponse == const.DELETE_IN_PROGRESS:

        print(const.ANIM_STRING[anim_index % const.ANIM_LEN], end="\r")   
        
        if running_time % const.DELAY_TIME_FOR_DESCRIBE_CHANGE_SET == 0.0:
            response = _session.cf.describe_stacks(StackName=name) 
            response = response[const.STACKS][0]
            eResponse = response[const.STACK_STATUS]
   
        anim_index += 1
        time.sleep(const.DELAY_TIME_FOR_ANIMATION)
        running_time += const.DELAY_TIME_FOR_ANIMATION

    return eResponse




__author__ = "Sivadon Chaisiri"
__copyright__ = "Copyright (c) 2020 Sivadon Chaisiri"
__license__ = "MIT License"


from setuptools import setup, find_packages

setup(
    name='pyhanga',
    packages=['pyhanga'],
    author='Sivadon Chaisiri',
    author_email='javaboom@gmail.com',    
    version='0.90',
    license='MIT',
    description = 'Custom Python-based CloudFormation Command Line Interface',
    url = 'https://github.com/boomary/pyhanga',   
    download_url = 'https://github.com/boomary/pyhanga/archive/v_90.tar.gz', 
    keywords = ['AWS', 'CloudFormation', 'CLI'],    
    install_requires=[
        'click',
        'boto3',
    ],
    include_package_data=True,   
    entry_points={
        'console_scripts': ['pyhanga=pyhanga.commands:cli']
    },
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3',
    ],
)



__author__ = "Sivadon Chaisiri"
__copyright__ = "Copyright (c) 2020 Sivadon Chaisiri"
__license__ = "MIT License"

import sys

import click
import boto3
import botocore

from . import _session
from . import pyhanga_constants as const
from . import pyhanga_util as util

@click.option('--field', '-f',
                help='Field to be printed out. You can print out one or more fields:\n'
                     'pyhanga list -f <field> -f <field> ...\n'
                     'By default, StackName and StackStatus are printed out.',
                default = tuple([const.STACK_NAME, const.STACK_STATUS]),
                multiple=True,
                type=click.Choice(const.STACK_SUMMARY_FILEDS, case_sensitive=False))
              

@click.option('--match-name', '-m',
                help='Search stacks based on a condition matching their name:\n'
                        'Format: pyhanga -m CHOICE VALUE\n'
                        'VALUE is the string to be searched.\n'
                        'CHOICE is a match condition type that can be contains, startswith, or endswith.\n'
                        'By default, all stack names are searched.',
                default=tuple([const.EXACTLY, const.ALL]),          
                nargs=2, type=click.Tuple([click.Choice(const.STRING_MATCH_CONDITIONS, case_sensitive=False), str]))

@click.option('--match-status', '-s',
                help='Search stacks based on a condition matching their status type\n'
                     'You can search one or more status types can be listed\n:'
                     'pyhanga list -s <type> --s <type> ...\n'
                     'By default, all status types except DELETE_COMPLETE are searched.',  
                default=const.DEFAULT_LSTACKSTATUS_FIELDS,
                multiple=True,
                type=click.Choice(const.STACK_STATUS_FILTERS, case_sensitive=False))


@click.command(name='list')
def list_stacks(field, match_name, match_status):
    """
    List stacks
    """
    _list_stacks(field, match_name, match_status)


def _list_stacks(field, match_name, match_status):
    mn_cond, mn_value = match_name

    try:
        response = _session.cf.list_stacks()
    except botocore.exceptions.ClientError as error:
        util.handleClientError(error)
    # except:
    #     click.secho(const.ERM_OTHERS, bg=const.BG_ERROR, fg=const.FG_ERROR)
    #     sys.exit(const.ERC_OTHERS)  

    count_results = 0
    json = response[const.STACK_SUMMARIES]
    match_status = util.uppercaseTuple(match_status)

    for response in json:

        stackStatus = str(response[const.STACK_STATUS])
        
        contOuterLoop = True
        for s in match_status:
            if stackStatus != s:
                continue
            else:
                contOuterLoop = False
                break
        
        if contOuterLoop:
            continue

        stackName = response[const.STACK_NAME]
        mn_cond = mn_cond.lower()
         
        if mn_cond == const.EXACTLY and mn_value == const.ALL:
            pass
        elif mn_cond == const.EXACTLY and mn_value != stackName:
            continue
        elif mn_cond == const.CONTAINS and mn_value not in stackName:
            continue
        elif mn_cond == const.STARTS_WITH and not stackName.startswith(mn_value):
            continue
        elif mn_cond == const.ENDS_WITH and not stackName.endswith(mn_value):
            continue  
        
        field = util.recaseTuple(field, const.STACK_SUMMARY_FILEDS)
        iField = iter(field)
        row = response[next(iField)]
        
        for i in iField:
            value = response.get(i)
            col = str(value) if value else const.NULL
            row = row + const.DELIM + col
        click.echo(row)
        count_results += 1
    
    click.secho('\nTotal number of stacks found:', fg=const.FG_INF)
    click.echo(count_results)


__author__ = "Sivadon Chaisiri"
__copyright__ = "Copyright (c) 2020 Sivadon Chaisiri"
__license__ = "MIT License"

import sys
import os
import random
import time

import click
import boto3
import botocore

from . import _session
from . import pyhanga_constants as const
from . import pyhanga_util as util
from . import upload_object

def deploy_stack(name, template, bucket, reuse, object_prefix, params, tags, upload, yes, iam, named_iam, default, isUpdated):

    if not reuse and not default and template is None:
        click.secho('Either --template (-t), or --default (-d), or --reuse (-r) option is required.', bg=const.BG_ERROR, fg=const.FG_ERROR)
        sys.exit(const.ERC_OTHERS)   

    if not reuse and bucket is None:
        click.secho('When --template (-t) or --default (-d) option is used, --bucket (-b) option is required.', bg=const.BG_ERROR, fg=const.FG_ERROR)
        sys.exit(const.ERC_OTHERS)  

    if not reuse and default and (template is not None or params is not None or tags is not None):
        click.secho('The --default (-d) option cannot be used with template, params, or tags option.', bg=const.BG_ERROR, fg=const.FG_ERROR)
        sys.exit(const.ERC_OTHERS)   

    if reuse and (bucket is not None or upload or template is not None):
        click.secho('The --reuse (-r) option cannot be used with --template (-t), --bucket (-b), or --upload (-u) option.', bg=const.BG_ERROR, fg=const.FG_ERROR)
        sys.exit(const.ERC_OTHERS)     
    
    if default:
        template = name + const.YAML_EXT
        params = name + '-params' + const.JSON_EXT
        tags = name + '-tags' + const.JSON_EXT    
        click.secho('Use the default file option.', fg=const.FG_INF)

    if not reuse:
        templateUrl = _formUrl_and_upload(template, bucket, object_prefix, upload)

    if params is not None:
        baseName = os.path.basename(params)        
        paramList = util.getJsonDataFromFile(params, baseName)
        click.secho('Use the parameter file: %s' % baseName, fg=const.FG_INF)
    else:
        paramList = list()
         
    if tags is not None:
        baseName = os.path.basename(tags)          
        tagList = util.getJsonDataFromFile(tags, baseName)
        click.secho('Use the tag file: %s' % baseName, fg=const.FG_INF) 
    else:
        tagList = list()           
        
    try:
        cName = name + str(random.randint(100000,999999)) # change set name
        cType = 'UPDATE' if isUpdated else 'CREATE'

        capabilities = []
        if iam:
            capabilities.append(const.CAPABILITY_IAM)
        if named_iam:
            capabilities.append(const.CAPABILITY_NAMED_IAM)

        if not reuse:
            response = _session.cf.create_change_set(StackName=name,
                                                TemplateURL=templateUrl,
                                                Capabilities=capabilities,
                                                Parameters=paramList,
                                                ChangeSetName=cName,
                                                ChangeSetType=cType,
                                                Tags=tagList)
        else:
            response = _session.cf.create_change_set(StackName=name,
                                                UsePreviousTemplate=reuse,
                                                Capabilities=capabilities,
                                                Parameters=paramList,
                                                ChangeSetName=cName,
                                                ChangeSetType=cType,
                                                Tags=tagList)            
        csId = response[const.CHANGESET_ID]
        stackId = response[const.STACK_ID]
        click.secho('\nThe following change set (ARN - name) is being created:', fg=const.FG_INF)                                                
        click.echo('%s - %s' % (csId, cName))                                                     
        click.secho('and deployed to the stack:', fg=const.FG_INF)
        click.echo(stackId)  

        response = _wait_for_created_changeset(name, cName)
        # click.echo(response)
        csExecutionStatus = response[const.CS_EXECUTION_STATUS]

        if csExecutionStatus == const.CS_AVAILABLE:
            click.secho('The change set is ready!', fg=const.FG_INF)

            # TODO: Change set results to be shown for confirmation
            _describe_changeset(name, cName, response)

            # If --yes flag is not specified, prompt the confirmation question.
            if not yes: 
                yes = util.query_yes_no('\nDo you want to execute this change set?', 'no')
            
            if not yes:
                if isUpdated:
                    _session.cf.delete_change_set(StackName=name,
                                                    ChangeSetName=cName)
                else:
                    _session.cf.delete_stack(StackName=name)
                click.secho('You have aborted this change set.', fg=const.FG_INF) 
            else:
                response = _session.cf.execute_change_set(StackName=name,
                                                    ChangeSetName=cName)  
                click.secho('The change set is being deployed.', fg=const.FG_INF)    
                eResponse = _wait_for_executed_changeset(name)  
                if (eResponse in [const.UPDATE_COMPLETE, const.IMPORT_COMPLETE, const.CREATE_COMPLETE]):
                    click.secho('\nThe executed change set was complete.', fg=const.FG_INF)
                else:
                    click.secho('\nThe executed change set got rollback with status: %s' % eResponse, fg=const.FG_ERROR)
        else:
            click.secho('Nothing to be created/updated.', fg=const.FG_INF) 

    except botocore.exceptions.ClientError as error:
        util.handleClientError(error)
    # except:
    #     click.secho(const.ERM_OTHERS, bg=const.BG_ERROR, fg=const.FG_ERROR)
    #     sys.exit(const.ERC_OTHERS)  


def _describe_changeset(name, cName, response):
    dictChanges = {const.CS_ADD: [],
                    const.CS_REMOVE: [], 
                    const.CS_MODIFY: []}
    click.echo()

    while True:
        changes = response[const.CS_CHANGES]
        for change in changes:
            rChange = change[const.CS_RESOURCE_CHANGE]
            action = rChange[const.CS_ACTION]
            resourceType = rChange[const.RESOURCE_TYPE]
            logicalResourceId = rChange[const.LOGICAL_RESOURCE_ID]
            replacement = rChange.get(const.CS_REPLACEMENT)  
            replacement = replacement if replacement is not None else 'False'
            dictChanges[action].append((resourceType, logicalResourceId, replacement))

        nextToken = response.get(const.NEXT_TOKEN)
        if nextToken is None:
            break
        else:
            response = _session.cf.describe_change_set(StackName=name,
                                                        ChangeSetName=cName,
                                                        NextToken=nextToken) 
    
    # Summary of resources
    lActions = [const.CS_ADD, const.CS_MODIFY, const.CS_REMOVE]
    pActions = ['added', 'modified', 'removed']
    sActions = ['+', '*', '-']
    for i in range(3):
        if len(dictChanges[lActions[i]]) > 0:
            sAction = sActions[i]
            click.secho('The following resources will be %s:' % pActions[i], fg=const.FG_INF) 
            for resource in dictChanges[lActions[i]]:
                resourceType, logicalResourceId, replacement = resource
                if replacement == 'False':
                    click.echo('  %s %s having ID: %s.' % (sAction, resourceType, logicalResourceId))
                elif replacement == 'True':
                    click.echo('  %s %s having ID: %s to be replaced.' % (sAction, resourceType, logicalResourceId))
                else:
                    click.echo('  %s %s having ID: %s to be conditionally replaced.' % (sAction, resourceType, logicalResourceId))


def _formUrl_and_upload(template, bucket, object_prefix, upload):
    object_prefix = util.reformS3Prefix(object_prefix)

    try:
        with open(template, 'r') as fTemplate:
            baseName = os.path.basename(fTemplate.name)
            object_key = object_prefix + baseName
            click.secho('Use the template file: %s' % baseName, fg=const.FG_INF)        
    except FileNotFoundError:
        util.handleFileNotFound(template)
    
    templateUrl = 'https://' + bucket + '.s3.amazonaws.com/' + object_key
    if (upload):
        upload_object._upload_object(bucket, object_key, os.path.abspath(template))
        click.secho('The template has been uploaded to the bucket.', fg=const.FG_INF)
    
    return templateUrl

def _wait_for_created_changeset(name, cName):   
    click.secho('Please wait while the change set is being created...', fg=const.FG_INF)

    eResponse = ''
    response = None   
    running_time = 0
    anim_index = 0
    
    while eResponse not in [const.CS_CREATE_COMPLETE, const.CS_FAILED]:
        print(const.ANIM_STRING[anim_index % const.ANIM_LEN], end="\r")   
        
        if running_time % const.DELAY_TIME_FOR_DESCRIBE_CHANGE_SET == 0.0:
            response = _session.cf.describe_change_set(StackName=name,
                                        ChangeSetName=cName) 
            eResponse = response[const.CS_STATUS]

        anim_index += 1
        time.sleep(const.DELAY_TIME_FOR_ANIMATION)
        running_time += const.DELAY_TIME_FOR_ANIMATION
    return response

def _wait_for_executed_changeset(name):  
    click.secho('Please wait while the change set is being executed...', fg=const.FG_INF)

    eResponse = ''
    response = None   
    running_time = 0
    anim_index = 0
    
    while eResponse not in [const.CREATE_COMPLETE,
                            const.UPDATE_COMPLETE,
                            const.DELETE_COMPLETE,
                            const.UPDATE_ROLLBACK_COMPLETE,
                            const.ROLLBACK_COMPLETE,
                            const.IMPORT_ROLLBACK_COMPLETE,
                            const.IMPORT_COMPLETE]:

        print(const.ANIM_STRING[anim_index % const.ANIM_LEN], end="\r")   
        
        if running_time % const.DELAY_TIME_FOR_DESCRIBE_CHANGE_SET == 0.0:
            response = _session.cf.describe_stacks(StackName=name) 
            response = response[const.STACKS][0]
            eResponse = response[const.STACK_STATUS]
   
        anim_index += 1
        time.sleep(const.DELAY_TIME_FOR_ANIMATION)
        running_time += const.DELAY_TIME_FOR_ANIMATION

    return eResponse




__author__ = "Sivadon Chaisiri"
__copyright__ = "Copyright (c) 2020 Sivadon Chaisiri"
__license__ = "MIT License"

import sys
import os
import random
import time

import click
import boto3
import botocore

from . import _session
from . import pyhanga_constants as const
from . import pyhanga_util as util
from .deploy_stack import _formUrl_and_upload

@click.option('--template', '-t',
                help='Stack template',
                type=click.STRING)

@click.option('--bucket', '-b',
                help='S3 bucket',
                required=True)

@click.option('--object-prefix', '-o',
                help='S3 object prefix (i.e., directory)\n'
                        'By default, the prefix is empty (i.e., root the bucket).',
                default='')

@click.option('--params', '--parameters', 
                help='Parameter file',
                default=None)

@click.option('--upload', '-u',
                help='Upload the template file to the bucket prefix',
                default=False,
                is_flag=True)  

@click.command(name='cost')
def cost_template(template, bucket, object_prefix, params, upload):
    """
    Estimate monthly cost of a stack
    """  

    templateUrl = _formUrl_and_upload(template, bucket, object_prefix, upload)
    if params is not None:
        baseName = os.path.basename(params)        
        paramList = util.getJsonDataFromFile(params, baseName)
        click.secho('Use the parameter file: %s' % baseName, fg=const.FG_INF)
    else:
        paramList = list()

    try:
        eResponse = _session.cf.estimate_template_cost(TemplateURL=templateUrl,
                                                        Parameters=paramList)  
        url = eResponse.get(const.URL)
        click.secho('\nThe estimated montly cost can be found at the following URL:', fg=const.FG_INF)              
        click.echo(url)
    except botocore.exceptions.ClientError as error:
        util.handleClientError(error)

